import os
import re
from spdx_license_list import LICENSES
from scancode.api import get_licenses
from typing import Iterable, List, Optional
from gimie.io import Resource
import tempfile

SPDX_IDS = list(LICENSES.keys())


def get_license_url(license_file: Resource) -> Optional[str]:
    """Takes a file-like resource containing a license text, and matches its content
    using the scancode API to get possible license matches. The best match is
    then returned as a spdx license URL.

    Examples
    --------
    >>> from gimie.io import LocalResource
    >>> get_license_url(LocalResource('LICENSE'))
    'https://spdx.org/licenses/Apache-2.0.html'
    """
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    temp_file.write(license_file.open().read())
    temp_file.close()

    license_detections = get_licenses(temp_file.name, include_text=True)[
        "license_detections"
    ]
    license_id = get_license_with_highest_coverage(license_detections)  # type: ignore
    if license_id is None:
        return None
    spdx_license_id = get_spdx_license_id(license_id)
    os.remove(temp_file.name)
    if spdx_license_id:
        return f"https://spdx.org/licenses/{str(spdx_license_id)}.html"

    return None


def get_spdx_license_id(
    license_id: str,
    spdx_ids: Iterable[str] = SPDX_IDS,
) -> Optional[str]:
    """Given a scancode API license ID also known as a license detection, returns the correctly capitalized
    spdx id corresponding to it.

    Parameters
    ----------
    license_id:
        A license id to match with SPDX licenses.
    spdx_ids:
        An iterable of (reference) SPDX license ids.

    Examples
    --------
    >>> get_spdx_license_id('apache-2.0')
    'Apache-2.0'
    >>> get_spdx_license_id('gpl-3.0')
    'GPL-3.0'
    """

    lower_spdx_ids = {spdx.lower(): spdx for spdx in spdx_ids}

    return lower_spdx_ids.get(license_id.lower(), None)


def is_license_path(filename: str) -> bool:
    """Given an input filename, returns a boolean indicating whether the filename path looks like a license.

    Parameters
    ----------
    filename:
        A filename to check.

    Examples
    --------
    >>> is_license_path('LICENSE.txt')
    True
    >>> is_license_path('LICENSE-APACHE')
    True
    >>> is_license_path('README.md')
    False
    """
    if filename.startswith("."):
        return False
    pattern = r".*(license(s)?.*|lizenz|reus(e|ing).*|copy(ing)?.*)(\.(txt|md|rst))?$"
    if re.match(pattern, filename, flags=re.IGNORECASE):
        return True
    return False


def get_license_with_highest_coverage(
    license_detections: List[dict],
) -> Optional[str]:
    """Filters a list of "license detections" (the output of scancode.api.get_licenses)
    to return the one with the highest match percentage.
    This is used to select among multiple license matches from a single file.

    Parameters
    ----------
    license_detections:
        A list of license detections, as returned by scancode.api.get_licenses.

    Examples
    --------
    >>> from scancode.api import get_licenses
    >>> license_detections = get_licenses('LICENSE')['license_detections']
    >>> get_license_with_highest_coverage(license_detections)
    'apache-2.0'
    """
    highest_coverage = 0.0
    highest_license = None

    for detection in license_detections:

        matches = detection["matches"] if "matches" in detection else []
        for match in matches:
            match_coverage = match.get("score", 0)
            if match_coverage > highest_coverage:
                highest_coverage = match_coverage
                highest_license = match.get("license_expression", None)
    return highest_license



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions used throughout gimie."""

from typing import List, Literal
from urllib.parse import urlparse

from gimie.graph.namespaces import GIMIE


def validate_url(url: str):
    """Checks if input is a valid URL.
    credits: https://stackoverflow.com/a/38020041

    Examples
    -------------
    >>> validate_url('/data/my_repo')
    False
    >>> validate_url(532)
    False
    >>> validate_url('https://www.github.com/SDSC-ORD/gimie')
    True
    >>> validate_url('github.com/SDSC-ORD/gimie')
    False
    """
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except AttributeError:
        return False


def generate_uri(ref: str):
    """Given a reference (e.g. commit sha), return a URI.

    Parameters
    ----------
    path:
        Path to the repository, either local or a URL.


    Returns
    -------
    fair_uri:
        A unique resource identifier (URI) for the repository path.

    Examples
    --------
    >>> generate_uri("abc")
    'https://sdsc-ord.github.io/gimie/abc'
    """
    return str(GIMIE[ref])



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract classes for gimie objects."""
from abc import ABC, abstractmethod
from typing import List, Optional


from rdflib import Graph
from urllib.parse import urlparse
from gimie.sources.common.license import get_license_url, is_license_path
from gimie.io import Resource


class Extractor(ABC):
    """Extractor is an Abstract Base Class. It is only meant
    to define a standard interface for all extractors.

    All subclasses must implement extract() and to_graph() methods
    they are free to override the default serialize() and jsonld()
    """

    def __init__(
        self,
        url: str,
        base_url: Optional[str] = None,
        local_path: Optional[str] = None,
    ):
        self.url = url
        self.base_url = base_url
        self.local_path = local_path

    @abstractmethod
    def extract(self):
        """Extract metadata"""
        ...

    @abstractmethod
    def to_graph(self) -> Graph:
        """Generate an RDF graph from the instance"""
        return Graph()

    def list_files(self) -> List[Resource]:
        """List all files in the repository HEAD."""
        ...

    def serialize(self, format: str = "ttl", **kwargs) -> str:
        """Serialize the RDF graph representing the instance."""
        return self.to_graph().serialize(format=format, **kwargs)  # type: ignore

    def jsonld(self) -> str:
        """Alias for jsonld serialization."""
        return self.serialize(format="json-ld")

    @property
    def _id(self) -> str:
        """Unique identifier for the repository."""
        return self.url

    @property
    def path(self) -> str:
        """Path to the repository without the base URL."""
        if self.base_url is None:
            return urlparse(self.url).path.strip("/")
        return self.url.removeprefix(self.base_url).strip("/")

    @property
    def base(self) -> str:
        """Base URL of the remote."""
        if self.base_url is None:
            url = urlparse(self.url)
            return f"{url.scheme}://{url.netloc}"
        return self.base_url

    def _get_licenses(self) -> List[str]:
        """Extracts SPDX License URLs from the repository."""
        license_files = filter(
            lambda p: is_license_path(p.name), self.list_files()
        )
        license_urls = []
        for file in license_files:
            license_url = get_license_url(file)
            if license_url:
                license_urls.append(license_url)
        return license_urls



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Orchestration of multiple extractors for a given project.
This is the main entry point for end-to-end analysis."""
from tempfile import gettempdir, TemporaryDirectory
from typing import Iterable, List, Optional, Tuple, Union

import shutil
import git
from rdflib import Graph
from urllib.parse import urlparse

from gimie.graph.operations import combine_graphs
from gimie.utils import validate_url
from gimie.sources import SOURCES
from gimie.sources.abstract import Extractor
from gimie.utils import validate_url


class Project:
    """A class to represent a project's git repository.


    Parameters
    ----------
    path:
        The full path (URL) of the repository.
    base_url:
        The base URL of the git remote. Can be used to
        specify delimitation between base URL and project name.
    sources:
        The metadata sources to use.

    Examples
    --------
    >>> proj = Project("https://github.com/SDSC-ORD/gimie")
    """

    def __init__(
        self,
        path: str,
        base_url: Optional[str] = None,
        sources: Optional[Union[Iterable[str], str]] = None,
    ):

        sources = normalize_sources(path, sources)
        self.base_url = base_url
        self.project_dir = None
        self._cloned = False
        if validate_url(path):
            self.url = path
            # We only need to clone a remote project
            # if a local extractor is enabled
            if any(map(is_local_source, sources)):
                self.project_dir = self.clone(path)
        else:
            self.project_dir = path

        self.extractors = self.get_extractors(sources)
        for ex in self.extractors:
            ex.extract()

    def clone(self, url: str) -> str:
        """Clone target url in a new temporary directory"""
        target_dir = TemporaryDirectory().name
        cloned = git.Repo.clone_from(url, target_dir)  # type: ignore
        self._cloned = True

        return str(cloned.working_tree_dir)

    def get_extractors(self, sources: Iterable[str]) -> List[Extractor]:

        extractors: List[Extractor] = []
        for src in sources:
            extractor = get_extractor(
                self.url,
                src,
                base_url=self.base_url,
                local_path=self.project_dir,
            )
            extractors.append(extractor)

        return extractors

    def to_graph(self) -> Graph:
        graphs = map(lambda ex: ex.to_graph(), self.extractors)
        combined_graph = combine_graphs(*graphs)
        return combined_graph

    def serialize(self, format: str = "ttl", **kwargs):
        return self.to_graph().serialize(format=format, **kwargs)

    def cleanup(self):
        """Recursively delete the project. Only works
        for remote (i.e. cloned) projects."""
        try:
            tempdir = gettempdir()
            if self.project_dir is not None:
                in_temp = self.project_dir.startswith(tempdir)
                if self._cloned and in_temp:
                    shutil.rmtree(self.project_dir)
        except AttributeError:
            pass

    def __del__(self):
        self.cleanup()


def split_git_url(url) -> Tuple[str, str]:
    base_url = urlparse(url).scheme + "://" + urlparse(url).netloc
    project = urlparse(url).path.strip("/")
    return base_url, project


def normalize_sources(
    path: str, sources: Optional[Union[Iterable[str], str]] = None
) -> List[str]:
    """Input validation and normalization for metadata sources.
    Returns a list of all input sources.

    Parameters
    ----------
    path :
        The path to the repository, either a local path or a URL.
    sources:
        The metadata sources to use. If None only the git provider is used.

    Returns
    -------
    List[str]
        A list of all input sources. If the git provider was missing
        from input sources, it is inferred from path.

    Examples
    --------
    >>> normalize_sources("https://github.com/SDSC-ORD/gimie")
    ['github']
    >>> normalize_sources("https://github.com/SDSC-ORD/gimie", "github")
    ['github']
    >>> normalize_sources("https://github.com/SDSC-ORD/gimie", ["github"])
    ['github']
    """
    if isinstance(sources, str):
        norm = [sources]
    elif isinstance(sources, Iterable):
        norm = list(sources)
    elif sources is None:
        norm = []
    else:
        raise ValueError(f"Invalid sources: {sources}")

    git_provider = get_git_provider(path)
    if git_provider not in norm:
        norm.append(git_provider)

    if (not validate_url(path)) and any(map(is_remote_source, norm)):
        raise ValueError("Cannot use a remote source with a local project.")
    return norm


def get_extractor(
    url: str,
    source: str,
    base_url: Optional[str] = None,
    local_path: Optional[str] = None,
) -> Extractor:
    """Instantiate the correct extractor for a given source.

    Parameters
    -----------
    URL
        Where the repository metadata is extracted from.
    source
        The source of the repository (git, gitlab, github, ...).
    base_url
        The base URL of the git remote.
    local_path
        If applicable, the path to the directory where the
        repository is cloned.
    """
    return SOURCES[source].extractor(
        url, base_url=base_url, local_path=local_path
    )


def is_valid_source(source: str) -> bool:
    """Check if input is a valid source for gimie."""
    return source in SOURCES


def is_remote_source(source: str) -> bool:
    """Check if input is a valid remote source for gimie."""
    if is_valid_source(source):
        return SOURCES[source].remote
    return False


def is_local_source(source: str) -> bool:
    """Check if input is a valid local source for gimie."""
    return not is_remote_source(source)


def is_git_provider(source: str) -> bool:
    """Check if input is a valid git provider for gimie."""
    if is_valid_source(source):
        return SOURCES[source].git
    return False


def get_git_provider(url: str) -> str:
    """Given a git repository URL, return the corresponding git provider.
    Local path or unsupported git providers will return "git"."""
    # NOTE: We just check if the provider name is in the URL.
    # We may want to use a more robust check.
    if validate_url(url):
        for name, prov in SOURCES.items():
            if prov.git and prov.remote and name in url:
                return name
    # Fall back to local git if local path of unsupported provider
    return "git"



# Gimie
# Copyright 2022 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import tempfile
from dataclasses import dataclass
from datetime import datetime
from dateutil.parser import isoparse
from functools import cached_property
import os
import requests
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urlparse
from dotenv import load_dotenv
from calamus import fields
from calamus.schema import JsonLDSchema
from rdflib import Graph

from gimie.sources.abstract import Extractor
from gimie.models import (
    Organization,
    OrganizationSchema,
    Person,
    PersonSchema,
)
from gimie.graph.namespaces import SDO

from gimie.io import RemoteResource
from gimie.sources.common.queries import (
    send_rest_query,
    send_graphql_query,
)

GH_API = "https://api.github.com"
load_dotenv()


def query_contributors(
    url: str, headers: Dict[str, str]
) -> List[Dict[str, Any]]:
    """Queries the list of contributors of target repository
    using GitHub's REST and GraphQL APIs. Returns a list of GraphQL User nodes.
    NOTE: This is a workaround for the lack of a contributors field in the GraphQL API."""
    owner, name = urlparse(url).path.strip("/").split("/")
    # Get contributors (available in the REST API but not GraphQL)
    data = f"repos/{owner}/{name}/contributors"
    contributors = send_rest_query(GH_API, data, headers=headers)
    ids = [contributor["node_id"] for contributor in contributors]
    # Get all contributors' metadata in 1 GraphQL query
    users_query = """
    query users($ids: [ID!]!) {
        nodes(ids: $ids) {
            ... on User {
                avatarUrl
                company
                login
                name
                organizations(first: 100) {
                    nodes {
                        avatarUrl
                        description
                        login
                        name
                        url
                    }
                }
                url
            }
        }
    }"""

    contributors = send_graphql_query(
        GH_API, users_query, data={"ids": ids}, headers=headers
    )
    # Drop empty users (e.g. dependabot)
    return [user for user in contributors["data"]["nodes"] if user]


@dataclass
class GithubExtractor(Extractor):
    """Extractor for GitHub repositories. Uses the GitHub GraphQL API to
    extract metadata into linked data.
    url: str
        The url of the git repository.
    base_url: Optional[str]
        The base url of the git remote.
    """

    url: str
    base_url: Optional[str] = None
    local_path: Optional[str] = None

    token: Optional[str] = None
    author: Optional[Union[Organization, Person]] = None
    contributors: Optional[List[Person]] = None
    prog_langs: Optional[List[str]] = None
    download_url: Optional[str] = None
    description: Optional[str] = None
    date_created: Optional[datetime] = None
    date_modified: Optional[datetime] = None
    date_published: Optional[datetime] = None
    parent_repository: Optional[str] = None
    keywords: Optional[List[str]] = None
    license: Optional[List[str]] = None
    software_version: Optional[str] = None

    def to_graph(self) -> Graph:
        """Convert repository to RDF graph."""
        jd = GithubExtractorSchema().dumps(self)
        g: Graph = Graph().parse(format="json-ld", data=str(jd))
        g.bind("schema", SDO)
        return g

    def list_files(self) -> List[RemoteResource]:
        """takes the root repository folder and returns the list of files present"""
        file_list = []
        file_dict = self._repo_data["object"]["entries"]
        repo_url = self._repo_data["url"]
        defaultbranchref = self._repo_data["defaultBranchRef"]["name"]

        for item in file_dict:
            file = RemoteResource(
                name=item["name"],
                url=f'{repo_url}/blob/{defaultbranchref}/{item["path"]}',
                headers=self._set_auth(),
            )
            file_list.append(file)
        return file_list

    def extract(self):
        """Extract metadata from target GitHub repository."""
        data = self._repo_data
        self.author = self._get_author(data["owner"])
        self.contributors = self._fetch_contributors()
        self.description = data["description"]
        self.date_created = isoparse(data["createdAt"][:-1])
        self.date_modified = isoparse(data["updatedAt"][:-1])
        if data["parent"]:
            self.parent_repository = data["parent"]["url"]
        if data["latestRelease"]:
            self.date_published = isoparse(
                data["latestRelease"]["publishedAt"]
            )

        self.license = self._get_licenses()
        if data["primaryLanguage"] is not None:
            self.prog_langs = [data["primaryLanguage"]["name"]]
        self.keywords = self._get_keywords(*data["repositoryTopics"]["nodes"])
        last_release = data["latestRelease"]
        if last_release is not None:
            self.version = last_release["name"]
            self.download_url = (
                f"{self.url}/archive/refs/tags/{self.version}.tar.gz"
            )

    @cached_property
    def _repo_data(self) -> Dict[str, Any]:
        """Repository metadata fetched from GraphQL endpoint."""
        owner, name = self.path.split("/")
        data = {"owner": owner, "name": name}
        repo_query = """
        query repo($owner: String!, $name: String!) {
            repository(name: $name, owner: $owner) {
                url
                parent {url}
                createdAt
                description
                latestRelease {
                    publishedAt
                    name
                }
                defaultBranchRef {
                    name
                }
                object(expression: "HEAD:") {
                    ... on Tree {

                        entries {
                            name
                            path
                            }
                        }
                    }
                mentionableUsers(first: 100) {
                    nodes {
                        login
                        name
                        avatarUrl
                        company
                        organizations(first: 100) {
                            nodes {
                                avatarUrl
                                description
                                login
                                name
                                url
                            }
                        }
                        url
                    }
                }
                name
                owner {
                    avatarUrl
                    login
                    url
                    ... on User {
                        company
                        name
                        organizations(first: 100) {
                            nodes {
                                avatarUrl
                                description
                                login
                                name
                                url
                            }
                        }
                    }
                    ... on Organization {
                        name
                        description
                    }
                }
                primaryLanguage {
                    name
                }
                repositoryTopics(first: 10) {
                    nodes {
                        topic {
                            name
                        }
                    }
                }
                updatedAt
                url
            }
        }
        """
        response = send_graphql_query(
            GH_API, repo_query, data, self._set_auth()
        )

        if "errors" in response:
            raise ValueError(response["errors"])

        return response["data"]["repository"]

    def _fetch_contributors(self) -> List[Person]:
        """Queries the GitHub GraphQL API to extract contributors through the commit list.
        NOTE: This is a workaround for the lack of a contributors field in the GraphQL API."""
        headers = self._set_auth()
        contributors = []
        resp = query_contributors(self.url, headers)
        for user in resp:
            contributors.append(self._get_user(user))
        return list(contributors)

    def _set_auth(self) -> Any:
        """Set authentication headers for GitHub API requests."""
        try:
            if not self.token:
                self.token = os.environ.get("GITHUB_TOKEN")
                assert self.token
            headers = {"Authorization": f"token {self.token}"}

            login = requests.get(f"{GH_API}/user", headers=headers)
            assert login.json().get("login")
        except AssertionError:
            return {}
        else:
            return headers

    def _get_keywords(self, *nodes: Dict[str, Any]) -> List[str]:
        """Extract names from GraphQL topic nodes."""
        return [node["topic"]["name"] for node in nodes]

    def _get_organization(self, node: Dict[str, Any]) -> Organization:
        """Extract details from a GraphQL organization node."""
        return Organization(
            _id=node["url"],
            name=node["login"],
            description=node["description"],
            legal_name=node["name"],
            logo=node["avatarUrl"],
        )

    def _get_author(self, node: Dict[str, Any]) -> Union[Organization, Person]:
        """Given the GraphQL node for a repository owner,
        return the author as a Person or Organization object."""

        if "organizations" in node:
            return self._get_user(node)

        return self._get_organization(node)

    def _get_user(self, node: Dict[str, Any]) -> Person:
        """Extract details from a GraphQL user node."""
        # Get user's affiliations
        orgs = [
            self._get_organization(org)
            for org in node["organizations"]["nodes"]
        ]
        return Person(
            _id=node["url"],
            identifier=node["login"],
            name=node["name"],
            affiliations=orgs,
        )


class GithubExtractorSchema(JsonLDSchema):
    """This defines the schema used for json-ld serialization."""

    _id = fields.Id()
    path = fields.String(SDO.name)
    author = fields.Nested(SDO.author, [PersonSchema, OrganizationSchema])
    contributors = fields.Nested(SDO.contributor, PersonSchema, many=True)
    prog_langs = fields.List(SDO.programmingLanguage, fields.String)
    download_url = fields.Raw(SDO.downloadUrl)
    description = fields.String(SDO.description)
    date_created = fields.Date(SDO.dateCreated)
    date_modified = fields.Date(SDO.dateModified)
    date_published = fields.Date(SDO.datePublished)
    license = fields.List(SDO.license, fields.IRI)
    url = fields.IRI(SDO.codeRepository)
    # NOTE: parent_repository is not available for GitLab's GraphQL API in 2023. Add for GitLab when available
    parent_repository = fields.IRI(SDO.isBasedOn)
    keywords = fields.List(SDO.keywords, fields.String)
    version = fields.String(SDO.version)

    class Meta:
        rdf_type = SDO.SoftwareSourceCode
        model = GithubExtractor
        add_value_types = False



"""Standard input interfaces to local or remote resources for gimie."""

import io
import os
from pathlib import Path
import requests
from typing import Optional, Union


class Resource:
    """Abstract class for buffered read-only access to local or remote resources via
    a file-like interface."""

    name: str

    def open(self) -> io.BufferedReader:
        raise NotImplementedError


class LocalResource(Resource):
    """Providing buffered read-only access to local data.

    Parameters
    ----------
    name: the name of the resource, typically the filename.
    url: the URL where the resource. can be downladed from.
    headers: optional headers to pass to the request.

    Examples
    --------
    >>> from gimie.io import LocalResource
    >>> resource = LocalResource("README.md")
    """

    def __init__(self, path: Union[str, os.PathLike]):
        self.path: Path = Path(path)

    def open(self, mode="r") -> io.BufferedReader:
        return io.BufferedReader(io.FileIO(self.path, mode))

    @property
    def name(self) -> str:
        return self.path.name


class RemoteResource(Resource):
    """Provides buffered read-only access to remote data.

    Parameters
    ----------
    name: the name of the resource, typically the filename.
    url: the URL where the resource. can be downladed from.
    headers: optional headers to pass to the request.

    Examples
    --------
    >>> from gimie.io import RemoteResource
    >>> url = "https://raw.githubusercontent.com/SDSC-ORD/gimie/main/README.md"
    >>> resource = RemoteResource("README.md", url)
    """

    def __init__(self, name: str, url: str, headers: Optional[dict] = None):
        self.name = name
        self.url = url
        self.headers = headers or {}

    def open(self) -> io.BufferedReader:
        resp = requests.get(
            self.url, headers=self.headers, stream=True
        ).iter_content(chunk_size=128)
        return iterable_to_stream(resp)


def iterable_to_stream(
    iterable, buffer_size=io.DEFAULT_BUFFER_SIZE
) -> io.BufferedReader:
    """
    Converts an iterable yielding bytestrings to a read-only input stream.
    Lets you use an iterable (e.g. a generator) that yields bytestrings as a read-only
    input stream.

    The stream implements Python 3's newer I/O API (available in Python 2's io module).
    For efficiency, the stream is buffered.

    credits: https://stackoverflow.com/a/20260030/8440675
    """

    class IterStream(io.RawIOBase):
        def __init__(self):
            self.leftover = ""

        def readable(self):
            return True

        def readinto(self, b):
            try:
                l = len(b)  # We're supposed to return at most this much
                chunk = self.leftover or next(iterable)
                output, self.leftover = chunk[:l], chunk[l:]
                b[: len(output)] = output
                return len(output)
            except StopIteration:
                return 0  # indicate EOF

    return io.BufferedReader(IterStream(), buffer_size=buffer_size)
